//======================================================= file = inport.c =====
//=  VOQ input port for VOQ switch simulation model                           =
//=============================================================================
//=  Notes: 1) This is for PPVOQ with buffered crossbar model                 =
//=---------------------------------------------------------------------------=
//=  Contains: voq_arbit_rr() - VOQ arbitration by RR                         =
//=            voq_arbit_lqf() - VOQ aribitration by LQF                      =
//=            voq_arbit_ocf() - VOQ arbitration by OCF                       =
//=            inport() - Input port containing VOQ buffers                   =
//=---------------------------------------------------------------------------=
//=  History:  KJC (08/12/01) - Genesis                                       =
//=============================================================================
#define EXTERNAL        // This is an external file (used in common.h)

//----- Include files ---------------------------------------------------------
#include <assert.h>     // Needed for assert() macro
#include "csim.h"       // Needed for CSIM
#include "common.h"     // Needed for common stuff

//=============================================================================
//==  Poller for input port VOQ buffers                                      ==
//==   - RR (next poll starts one index past selected queue to tie break)    ==
//==   - Set Burst = 1 to disable thresholding feature                       ==
//=============================================================================
void voq_arbit_rr(int in_id)
{
  int      out_id;              // Output port id
  int      out_poll;            // Output port id to be polled
  int      flag;                // Flag variable
  int      i;                   // Loop counter

  create("voq_arbit_rr");

  // Poll input port VOQ buffers forever
  out_id = 0;
  Count[in_id] = Burst - 1;
  while(1)
  {
    // Test if any VOQ buffer transfers can be done
    flag = 0;
    for (i=0; i<N; i++)
      if ((queue_cnt(VOQ_first[in_id][i]) > 0) &&
          (queue_cnt(CP_event[in_id][i]) == 0))
        flag = 1;

    // If no transfer can be done, wait for an input port state change
    if (flag == 0) wait(State_change[in_id]);

    // RR poll to try to find the first VOQ that can be transferred
    for (i=0; i<N; i++)
    {
      out_poll = (out_id + i) % N;
      if ((queue_cnt(VOQ_first[in_id][out_poll]) > 0) &&
          (queue_cnt(CP_event[in_id][out_poll]) == 0))
      {
        set(VOQ_first[in_id][out_poll]);
        wait(Xfer_voq[in_id]);
        hold(0.0);
        if (((queue_cnt(VOQ_event[in_id][out_poll]) +
              queue_cnt(VOQ_first[in_id][out_poll])) > Thresh) &&
             (Count[in_id] > 0))
        {
          out_id = out_id;
          Count[in_id]--;
        }
        else
        {
          out_id = (out_poll + 1) % N;
          Count[in_id] = Burst - 1;
        }
        break;
      }
    }
  }
}

//=============================================================================
//==  Poller for input port VOQ buffers                                      ==
//==   - LQF (next poll starts one index past tied longest queue)            ==
//=============================================================================
void voq_arbit_lqf(int in_id)
{
  int      out_id;              // Output port id
  int      out_poll;            // Output port id to be polled
  int      longest;             // Longest VOQ length
  int      flag;                // Flag variable
  int      i;                   // Loop counter

  create("voq_arbit_lqf");

  // Poll input port VOQ buffers forever
  out_id = 0;
  while(1)
  {
    // Test if any VOQ buffer transfers can be done
    flag = 0;
    for (i=0; i<N; i++)
      if ((queue_cnt(VOQ_first[in_id][i]) > 0) &&
          (queue_cnt(CP_event[in_id][i]) == 0))
        flag = 1;

    // If no transfer can be done, wait for an input port state change
    if (flag == 0) wait(State_change[in_id]);

    // Poll to try to find longest queue
    longest = 0;
    for (i=0; i<N; i++)
    {
      out_poll = (out_id + i) % N;
      if ((queue_cnt(VOQ_first[in_id][out_poll]) > 0) &&
          (queue_cnt(CP_event[in_id][out_poll]) == 0))
      {
        if ((queue_cnt(VOQ_event[in_id][out_poll]) +
             queue_cnt(VOQ_first[in_id][out_poll])) > longest)
        {
          longest = queue_cnt(VOQ_event[in_id][out_poll]) +
                    queue_cnt(VOQ_first[in_id][out_poll]);
          out_id = out_poll;
        }
      }
    }

    // If an eligible VOQ was found, transfer the VOQ contents to the CP buffer
    if (longest > 0)
    {
      set(VOQ_first[in_id][out_id]);
      wait(Xfer_voq[in_id]);
      hold(0.0);
      out_id = (out_id + 1) % N;
    }
  }
}

//=============================================================================
//==  Poller for input port VOQ buffers                                      ==
//==   - OCF (uses floating point time-stamp comparisons)                    ==
//=============================================================================
void voq_arbit_ocf(int in_id)
{
  double   oldest;              // Oldest timestamp value
  int      out_id;              // Output port id
  int      flag;                // Flag variable
  int      i;                   // Loop counter

  create("voq_arbit_ocf");

  // Poll input port VOQ buffers forever
  while(1)
  {
    // Test if any VOQ buffer transfers can be done
    flag = 0;
    for (i=0; i<N; i++)
      if ((queue_cnt(VOQ_first[in_id][i]) > 0) &&
          (queue_cnt(CP_event[in_id][i]) == 0))
        flag = 1;

    // If no transfer can be done, wait for an input port state change
    if (flag == 0) wait(State_change[in_id]);

    // Poll to try to find oldest cell
    oldest = clock;
    out_id = -1;
    for (i=0; i<N; i++)
    {
      if ((queue_cnt(VOQ_first[in_id][i]) > 0) &&
          (queue_cnt(CP_event[in_id][i]) == 0))
      {
        if (Time_stamp[in_id][i] <= oldest)
        {
          oldest = Time_stamp[in_id][i];
          out_id = i;
        }
      }
    }

    // If an eligible VOQ was found, transfer the VOQ contents to the CP buffer
    if (out_id >= 0)
    {
      set(VOQ_first[in_id][out_id]);
      wait(Xfer_voq[in_id]);
      hold(0.0);
    }
  }
}

//=============================================================================
//==  Input port containing VOQ buffers                                      ==
//==   - Modified for OCF support to support HOL timestamp                   ==
//=============================================================================
void inport(int in_id, int out_id, double org_time)
{
  create("inport");

  // If VOQ is empty then set the arrival state change event
  if (queue_cnt(VOQ_first[in_id][out_id]) == 0) set(State_change[in_id]);

  // Queue the arriving cell in the appropriate VOQ buffer
  queue(VOQ_event[in_id][out_id]);
  assert(queue_cnt(VOQ_first[in_id][out_id]) == 0);
  Time_stamp[in_id][out_id] = org_time;
  queue(VOQ_first[in_id][out_id]);
  set(VOQ_event[in_id][out_id]);

  // Forward the cell to the crosspoint buffer (as "owned" by output port)
  assert(queue_cnt(CP_event[in_id][out_id]) == 0);
  assert(status(Bus[in_id]) == FREE);
  reserve(Bus[in_id]);
  hold(CELL_TIME);
  release(Bus[in_id]);
  out[out_id](in_id, out_id, org_time);

  // Set appropriate events
  set(Xfer_voq[in_id]);
  set(Xfer_cp[out_id]);

  return;
}
